/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/*!
@file corners.cpp
@ingroup ExamplePrograms
@brief Obtains <a href="http://en.wikipedia.org/wiki/Corner_detection#The_Harris_.26_Stephens_.2F_Plessey_corner_detection_algorithm">Harris corners</a> from the frames of a video input source.

@todo Detailed description should follow.

@section Usage
Once compiled, the program should be used with this command line:
\code
	./corners --URL=<path or url to video>
\endcode

En mathematica:
H = {{h1, h2, h3}, {h4, h5, h6}, {h7, h8, h9}};
Cross[{p, q, 1}, H .{x, y, 1}]

@param URL Video source identificator. See dynamic property URL for @ref QVMPlayerCamera.
@author PARP Research Group. University of Murcia, Spain.
*/

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <QDebug>
#include <math.h>


#include <qvcore/qvapplication.h>
#include <qvcameras/qvmplayercamera.h>
#include <qvgui/qvgui.h>

#include <qvdta/qvpolyline.h>
#include <qvdta/qvdta.h>

#include <TooN/TooN.h>
#include <TooN/numerics.h>	// Basic TooN library
#include <TooN/numhelpers.h>	// If you want a few extras
#include <TooN/SVD.h>		// If you want singular value decompositions
#include <TooN/LU.h>		// If you want lu triangular decompositions
#include <TooN/SymEigen.h>	// If you want Eigen decomposition of symmetric matrices

#define PI		3.14159
#define MAX(x,y)	(((x) > (y))?(x):(y))

Q_DECLARE_METATYPE(Matrix<>);

/// \cond
void GetHotPoints(const QVImage<sFloat> cornerResponseImage, QList<QPoint> &hotPoints, uInt maxWindow)
	{
	const uInt rows = cornerResponseImage.getRows(), cols = cornerResponseImage.getCols();

	QVImage<uChar> binaryCornerImage(cols, rows);
	FilterLocalMax(cornerResponseImage, binaryCornerImage, maxWindow, maxWindow);

	QVIMAGE_INIT_READ(uChar,binaryCornerImage);
	for (uInt row = 0; row < binaryCornerImage.getRows(); row++)
		for (uInt col = 0; col < binaryCornerImage.getCols(); col++)
			if (QVIMAGE_PIXEL(binaryCornerImage, col, row,0))
				hotPoints.append(QPoint(col, row));
	}

void GetMaximalPoints(const QVImage<sFloat> cornerResponseImage, QList<QPoint> &hotPoints, QList<QPoint> &maximalPoints, uInt maxPoints)
	{
	while( hotPoints.size() > 0 && maximalPoints.size() < maxPoints )
		{
		uInt maxIndex = 0;
		for (int n=0; n < hotPoints.size(); n++)
			if ( cornerResponseImage(hotPoints.at(n)) >  cornerResponseImage(hotPoints.at(maxIndex)) )
				maxIndex = n;

		maximalPoints.append(hotPoints.at(maxIndex));
		hotPoints.removeAt(maxIndex);
		}
	}

uInt	getClosestPointIndex(const QPoint point, const QList<QPoint> &pointList)
	{
	uInt index = 0;
	for (uInt n = 1; n < pointList.size(); n++)
		if ((point - pointList.at(n)).manhattanLength() < (point - pointList.at(index)).manhattanLength())
			index = n;

	return index;
	}

QPoint	getMeanPoint(const QList<QPoint> &pointList)
	{
	QPoint center(0,0);

	for (uInt n = 0; n < pointList.size(); n++)
		{
		center.rx() += pointList.at(n).x();
		center.ry() += pointList.at(n).y();
		}

	center.rx() = center.rx() / pointList.size();
	center.ry() = center.ry() / pointList.size();

	return center;
	}

double	angle(const QPoint &p)
	{
	double x = p.x(), y = p.y();
	if (x>0)
		if (y>=0)
			return atan(y/x);
		else
			return atan(y/x) + 2*PI;
	else if (x == 0)
		if (y>0)
			return PI/2;
		else
			return 3*PI/2;
	else // x < 0
		return atan(y/x)+PI;
	}

double	clockWiseAngle(const QPoint &p1, const QPoint &p2)
	{
	double clockAngle = angle(p2) - angle(p1);
	return (clockAngle < 0)? clockAngle + 2*PI:clockAngle;
	}

//////////////////////////////////////////
double	angle(const double x, const double y)
	{
	if (x>0)
		if (y>=0)
			return atan(y/x);
		else
			return atan(y/x) + 2*PI;
	else if (x == 0)
		if (y>0)
			return PI/2;
		else
			return 3*PI/2;
	else // x < 0
		return atan(y/x)+PI;
	}

double	clockWiseAngle(const double x1, const double y1, const double x2, const double y2)
	{
	double clockAngle = angle(x2,y2) - angle(x1,y1);
	return (clockAngle < 0)? clockAngle + 2*PI:clockAngle;
	}

// This function creates a list of the template points, first with the center point,
// then the rest of the points following a clockwise order around the center point.
QList<QPoint> GetTemplatePoints(uInt zoom, QPoint center)
	{
	QPoint rPoints[5] =
		//	Center		Norwest		Noreast		Southeast	Southwest
		{	QPoint(0,0),	QPoint(-1,-1),	QPoint(+1,-1),	QPoint(+1,+1),	QPoint(-1,+1)	};

	QList<QPoint> templatePoints;

	for (uInt i= 0; i < 5; i++)
		templatePoints.append(zoom*rPoints[i]+center);

	return templatePoints;
	}

// This is suposed to sort the points in a list, by... who knows?
// Anyway, it doesn't works
bool SortTemplatePoints(QList<QPoint> &points)
	{
	if (points.size() != 5)
		return false;

	QList<QPoint> result;

	// This array will keep the index of the points.
	uInt index[5];

	// Calculate the index of the central point
	uInt indexp = getClosestPointIndex(getMeanPoint(points), points);

	result.append(points.at(indexp));
	points.removeAt(indexp);

	// Calculate the index of the top left point
	double minDistance = 1000000;
	for (uInt n = 0; n < points.size(); n++)
		if ( points.at(n).manhattanLength() < minDistance )
			{
			minDistance = points.at(n).manhattanLength();
			indexp = n;
			}

	result.append(points.at(indexp));
	points.removeAt(indexp);

	// Calculate the other points
	while(points.size() > 0)
		{
		indexp = 0;
		double minAngle = clockWiseAngle( result.back() - result.front(), points.at(indexp) - result.front());
	
		for (uInt n = 1; n < points.size(); n++)
			{
			double actualAngle = clockWiseAngle( result.back() - result.front(), points.at(n) - result.front());
			if ( actualAngle < minAngle )
				{
				minAngle = actualAngle;
				indexp = n;
				}
			}

		result.append(points.at(indexp));
		points.removeAt(indexp);
		}

	points = result;

	return true;
	}


// This function returns the homography that maps the points from a source position to a
// destination position
Matrix<> CalibrateHomography(const Matrix<> &sourcePoints, const Matrix<> &destinationPoints)
	{
	Q_ASSERT(sourcePoints.num_cols() == 2);
	Q_ASSERT(sourcePoints.num_cols() == destinationPoints.num_cols());
	Q_ASSERT(sourcePoints.num_rows() == destinationPoints.num_rows());

	const uInt rows = sourcePoints.num_rows();

	// First, state the coefs matrix for the homogeneous linear system of equations, of the form:
	//	Ax = 0
	Matrix<> coefsMatrix(3*rows,9);

	//std::cout << "Points" << std::endl;
	for (uInt n = 0; n < rows; n++)
		{
		double	x = sourcePoints[n][0], y = sourcePoints[n][1],
			p = destinationPoints[n][0], q = destinationPoints[n][1];

		double	equation1[9] = {	0,	0,	0,	-x,	-y,	-1,	q*x,	q*y,	q},
			equation2[9] = {	x,	y,	1,	0,	0,	0,	-p*x,	-p*y,	-p},
			equation3[9] = {	-q*x,	-q*y,	-q,	p*x,	p*y,	p,	0,	0,	0};

		coefsMatrix[3*n] = Vector<9>(equation1);
		coefsMatrix[3*n+1] = Vector<9>(equation2);
		coefsMatrix[3*n+2] = Vector<9>(equation3);
		}

	// create the SVD of M
	SVD<> svdCoefsMatrix(coefsMatrix);

	Vector<9> x =  svdCoefsMatrix.get_VT()[8];
	Matrix<> homography(3,3);

	homography[0][0] = x[0];	homography[0][1] = x[1];	homography[0][2] = x[2];
	homography[1][0] = x[3];	homography[1][1] = x[4];	homography[1][2] = x[5];
	homography[2][0] = x[6];	homography[2][1] = x[7];	homography[2][2] = x[8];

	return homography;
	}

// This function maps a point given a planar homography
QPoint	applyHomography(const QPoint p, const Matrix<> homography)
	{
	const double values[3] = { p.x(), p.y(), 1 };
	Vector<3> p1 = Vector<3>(values), p2 = homography * p1;
	return QPoint(round(p2[0]/p2[2]), round(p2[1]/p2[2]));
	}


void normalizeHomogeneousCoordinates(Matrix<> &points)
	{
	const uInt cols = points.num_cols(), rows = points.num_rows();

	for (uInt i = 0; i < rows; i++)
		for (uInt j = 0; j < cols; j++)
			points[i][j] /= points[i][cols-1];
	}

void normalizeHomogeneousCoordinates(Vector<> &point)
	{
	for (uInt j = 0; j < 3; j++)
		point[j] /= point[2];
	}

// This function retunrs an integer value that indicates how well the homography maps the source points
// in a list, to the points in the destination list.
double	testErrorHomography(const Matrix<> &sourcePoints, const Matrix<> &destinationPoints, const Matrix<> homography)
	{
	const uInt cols = sourcePoints.num_cols(), rows = sourcePoints.num_rows();

	Matrix <> projectedPoints(sourcePoints.num_rows(),3);
	Matrix <> residuals (sourcePoints.num_rows(),3);

	projectedPoints = sourcePoints * homography.T();
	normalizeHomogeneousCoordinates(projectedPoints);
	residuals = projectedPoints - destinationPoints;

	double accum = 0;
	for (uInt i = 0; i < rows; i++)
		{
		double square = 0;
		for (uInt j = 0; j < cols; j++)
			square += residuals[i][j]*residuals[i][j];
		accum += sqrt(square);
		}

	return accum;
	}

// This function sustitudes warpPerspective from the qvipp, and applies the planar transformation H
// to a source image, storing the transformed image in the destination image
void	myWarpPerspective(const QVImage<uChar> &src, QVImage<uChar> &dest, const Matrix <> H)
	{
	const uInt cols = src.getCols(), rows = src.getRows();
	const QPoint center = QPoint(cols/2, rows/2);
	const Matrix<> Hinv = SVD<>(H).get_pinv();

	for (uInt col = 0; col < cols; col++)
		for (uInt row = 0; row < rows; row++)
			{
			QPoint p2 = QPoint(col, row), p1 = applyHomography(p2-center, Hinv);
			if (src.getROI().contains(p1))
				dest(p2) = src(p1);
			}
	}

// Given the four corners of a square, transformed by an homography, this function obtains two points
// of the line of the horizont.
/*bool getHorizont(const double x1, const double y1, const double x2, const double y2, const double x3, const double y3, const double x4, const double y4,
		double &x5, double &y5, double &x6, double &y6)
	{
	const double	x5num =		x1 *(x4 *(-y2 + y3) + x3 *(y2 - y4)) + x2 *(x4 *(y1 - y3) + x3 *(-y1 + y4)),
			x5denom =	x4 *(y1 - y2) + x3 *(-y1 + y2) + (x1 - x2)* (y3 - y4),
			y5num =		x4 *(y1 - y2)* y3 + x1* y2* y3 - x3* y1* y4 - x1* y2* y4 + x3 *y2* y4 + x2 *y1 *(-y3 + y4),
			y5denom =	x4 *(y1 - y2) + x3 *(-y1 + y2) + (x1 - x2)* (y3 - y4),

			x6num =		x3 *(x4 *(y1 - y2) + x1 *(y2 - y4)) + x2* (x4 *(-y1 + y3) + x1 *(-y3 + y4)),
			x6denom =	(x1 - x4) *(y2 - y3) + x3 *(y1 - y4) + x2 *(-y1 + y4),
			y6num =		-x2* y1* y3 + x4* y1* (-y2 + y3) + x3* y2* (y1 - y4) + x1* y2* y4 - x1* y3* y4 + x2* y3* y4,
			y6denom =	(x1 - x4)* (y2 - y3) + x3* (y1 - y4) + x2* (-y1 + y4);

	if (x5denom == 0 || y5denom == 0 || x6denom == 0 || y6denom == 0)
		return false;

	x5 = x5num / x5denom;
	y5 = y5num / y5denom;
	x6 = x6num / x6denom;
	y6 = y6num / y6denom;

	return true;
	}*/

Matrix <> RotationMatrix2d(const double theta)
	{
	const double m[3*3] =	{
				cos(theta),	-sin(theta),	0,
				sin(theta),	cos(theta),	0,
				0,		0,		1
				};

	return Matrix<3,3>(m);
	}

bool linesIntersectionPoint(const double x1, const double y1, const double x2, const double y2, const double x3, const double y3, const double x4, const double y4,
		double &x5, double &y5)
	{
	const double	x5num =		x1 *(x4 *(-y2 + y3) + x3 *(y2 - y4)) + x2 *(x4 *(y1 - y3) + x3 *(-y1 + y4)),
			x5denom =	x4 *(y1 - y2) + x3 *(-y1 + y2) + (x1 - x2)* (y3 - y4),
			y5num =		x4 *(y1 - y2)* y3 + x1* y2* y3 - x3* y1* y4 - x1* y2* y4 + x3 *y2* y4 + x2 *y1 *(-y3 + y4),
			y5denom =	x4 *(y1 - y2) + x3 *(-y1 + y2) + (x1 - x2)* (y3 - y4);

	if (x5denom == 0 || y5denom == 0)
		return false;

	x5 = x5num / x5denom;
	y5 = y5num / y5denom;

	return true;
	}

// This function gets the closest point in a line, to a given point, not contained in it
bool closestPointAtLine(const double xc, const double yc, const double x5, const double y5, const double x6, const double y6, double &x7, double &y7)
	{
	const double	x7num =		xc*x5*x5 + x6*(xc*x6-(yc-y5)*(y5-y6)) + x5*(-2*xc*x6+(yc-y6)*(y5-y6)),
			x7denom =	x5*x5-2*x5*x6+x6*x6+(y5-y6)*(y5-y6),
			y7num =		x6*x6*y5+yc*(y5-y6)*(y5-y6)+x5*x5*y6+xc*x6*(-y5+y6)+x5*((xc-x6)*y5-(xc+x6)*y6),
			y7denom =	x5*x5 - 2*x5*x6 + x6*x6 + (y5-y6)*(y5-y6);


	if (x7denom == 0 || y7denom == 0)
		return false;

	x7 = x7num / x7denom;
	y7 = y7num / y7denom;

	return true;

	}

// This function draws a line in an image, given two points of it.
void	myDrawLine(QVImage<uChar,3> &image, const int x1, const int y1, const int x2, const int y2)
	{
	const uInt rows = image.getRows(), cols = image.getCols();
	const double m = (double)(y2-y1)/(double)(x2-x1);

	for(int x = 0; x< cols; x++)
		{
		int y = m*(x - x1) + y1;
		if ( y >= 0 && y < rows)
			{
			image(x,y,0) = 255;
			image(x,y,1) = 0;
			image(x,y,2) = 0;
			}
		}
	}

// This function transforms a list of n points, to a nx3 matrix, containing the rows the
// homogeneous coordinates for each point in the list.
void pointListToMatrix(const QList<QPoint> &points, Matrix<> &matrix)
	{
	Q_ASSERT(points.size() == matrix.get_rows());
	Q_ASSERT(3 == matrix.get_cols());

	for (uInt n = 0; n < points.size(); n++)
		{
		double	v[3] = { points.at(n).x(), points.at(n).y(), 1 };
		matrix[n] = Vector<3>(v);
		}
	}


#define NORM(x,y)	(sqrt((x)*(x) + (y)*(y)))

void getSubpixelPrecissionPoints(const QVImage<sFloat> &weightImage, QList<QPoint> &pointList, Matrix<> &pointsMatrix, uInt margin)
	{
	const uInt	rows = weightImage.getRows(), cols = weightImage.getCols();

	QVIMAGE_INIT_READ(sFloat, weightImage);
	for (uInt n = 0; n < pointList.size(); n++)
		{
		const QPoint p = pointList.at(n);
		const int x = p.x(), y = p.y();
		if (x > margin && x < cols-margin && y > margin && y < rows-margin)
			{
			double v[3] = {0, 0, 1}, sum = 0;
			for (uint i = x-margin; i<=x+margin; i++)
				for (uint j = y-margin; j<=y+margin; j++)
					{
					const double factor = QVIMAGE_PIXEL(weightImage, i, j, 0);
					v[0] += i*factor;
					v[1] += j*factor;
					sum += factor;
					}

			v[0] = v[0] / sum;
			v[1] = v[1] / sum;

			pointsMatrix[n] = Vector<3>(v);
			}
		}
	}

class MyWorker: public QVWorker
	{
	public:
		MyWorker(QString name): QVWorker(name)
			{
			addProperty<double>("Max error", inputFlag, 2.5, "window size", 0, 10);
			addProperty<int>("Zoom", inputFlag, 20, "window size", 1, 256);
			addProperty<int>("Window size", inputFlag, 10, "window size", 1, 100);
			addProperty<int>("Margin", inputFlag, 1, "window size", 1, 64);
			addProperty<double>("Distance", inputFlag, 2, "window size", 0.1, 5);
			addProperty< QVImage<uChar,1> >("Input image", inputFlag|outputFlag);
			addProperty< QVImage<uChar,3> >("Corners", outputFlag);
			addProperty< QVImage<uChar,3> >("Wrapped", outputFlag);
			addProperty< QVImage<uChar,3> >("Wrapped2", outputFlag);
			addProperty< QVImage<sFloat,1> >("Response image", outputFlag);

			addTrigger("Grab homography");
			addProperty< Matrix<> >("Actual homography", outputFlag);
			}

		void processTrigger(QString triggerName)
			{
			Matrix <> H = getPropertyValue< Matrix<> >("Actual homography");
			std::cout <<	"H = {"
				<< "{" << H[0][0] << ", " << H[0][1] << ", " << H[0][2] << "},"
				<< "{" << H[1][0] << ", " << H[1][1] << ", " << H[1][2] << "},"
				<< "{" << H[2][0] << ", " << H[2][1] << ", " << H[2][2] << "}};" << std::endl;
			};

		void iterate()
			{
			const QVImage<uChar> image = getPropertyValue< QVImage<uChar,1> >("Input image");
			const uInt	rows = image.getRows(), cols = image.getCols(),
					sizeMax = getPropertyValue<int>("Window size"),
					zoom = getPropertyValue<int>("Zoom"),
					margin = getPropertyValue<int>("Margin");
					
			const double	maxError = getPropertyValue<double>("Max error"),
					distance = getPropertyValue<double>("Distance");

			QVImage<uChar,3> destino = image;

			timeFlag("grab Frame");

			////////////////////////////////////
			// Harris corner response image
			QVImage<sFloat>	temp(cols, rows), cornerResponseImage(cols, rows);
			HarrisCornerResponseImage(image, cornerResponseImage);
			//FilterGauss(temp, cornerResponseImage);

			timeFlag("Corner response image");

			//setPropertyValue< QVImage<sFloat> >("Response image", cornerResponseImage);
			//SobelCornerResponseImage(image, cornerResponseImage);

			////////////////////////////////////
			// Hot points
			QList<QPoint> hotPoints;
			GetHotPoints(cornerResponseImage, hotPoints, sizeMax);

			timeFlag("Get hotpoints");

			///////////////////////////////////////////
			// Calibración
			QList<QPoint> maximalPoints, templatePoints = GetTemplatePoints(1, QPoint(0, 0));
			GetMaximalPoints(cornerResponseImage, hotPoints, maximalPoints, 5);
			SortTemplatePoints(maximalPoints);

			timeFlag("Get max hotpoints");

			if (maximalPoints.size() == 5)
				{
				Matrix <> sourcePoints(5,3), sourcePoints2(5,3), destinationTemplatePoints(5,3);

				//QVImage<sFloat> temp(cols, rows);
				//FilterGauss(cornerResponseImage,temp);

				getSubpixelPrecissionPoints(cornerResponseImage, maximalPoints, sourcePoints,margin);
				pointListToMatrix(maximalPoints, sourcePoints2);
				pointListToMatrix(templatePoints, destinationTemplatePoints);

				/*const double meanColsRows = (double)(cols + rows)/2;

				for (uInt i = 0; i < 5; i++)
					{
					sourcePoints[i][0] = sourcePoints[i][0] / meanColsRows - 0.5;
					sourcePoints[i][0] = sourcePoints[i][0] / meanColsRows - 0.5;
					sourcePoints2[i][0] = sourcePoints[i][0] / meanColsRows - 0.5;
					sourcePoints2[i][0] = sourcePoints[i][0] / meanColsRows - 0.5;
					destinationTemplatePoints[i][0] /= 2;
					destinationTemplatePoints[i][0] /= 2;
					}*/

				Matrix<> H = CalibrateHomography(sourcePoints, destinationTemplatePoints);
				Matrix<> H2 = CalibrateHomography(sourcePoints2, destinationTemplatePoints);

				if (testErrorHomography(sourcePoints, destinationTemplatePoints, H) < maxError)
					{
					std::cout << "--------------------------" << std::endl;
					//std::cout << "points1" << sourcePoints << std::endl;
					//std::cout << "points2" << sourcePoints2 << std::endl;

					std::cout << "ERROR 1 " << testErrorHomography(sourcePoints, destinationTemplatePoints, H) << std::endl;
					std::cout << "ERROR 2 " << testErrorHomography(sourcePoints2, destinationTemplatePoints, H2) << std::endl;

					setPropertyValue< Matrix<> >("Actual homography", H);
	
					QVImage<uChar> wrapped(cols, rows);
					Set(wrapped,0);
	
					myWarpPerspective(image, wrapped, H);

					QVImage<uChar> wrapped2(cols, rows);
					Set(wrapped2,0);
	
					myWarpPerspective(image, wrapped2, H2);
	
					QVImage<uChar> diff(cols, rows);
					AbsDiff(wrapped, wrapped2, diff);
	
					// Ideas:
					// - Usar los puntos calibrados con precisión subpixel
					// - Usar puntos "sintéticos", a partir de los puntos de la plantilla y la matriz H
					//
					/*const double	x1 = maximalPoints.at(1).x(), y1 = maximalPoints.at(1).y(),
							x2 = maximalPoints.at(2).x(), y2 = maximalPoints.at(2).y(),
							x3 = maximalPoints.at(3).x(), y3 = maximalPoints.at(3).y(),
							x4 = maximalPoints.at(4).x(), y4 = maximalPoints.at(4).y();*/
					
					/*const double	x1 = sourcePoints[1][0], y1 = sourcePoints[1][1],
							x2 = sourcePoints[2][0], y2 = sourcePoints[2][1],
							x3 = sourcePoints[3][0], y3 = sourcePoints[3][1],
							x4 = sourcePoints[4][0], y4 = sourcePoints[4][1];*/

					//std::cout << "***\t" << x1 << ", " << y1 << "\t" << x2 << ", " << y2 << "\t" << x3 << ", " << y3 << "\t" << x4 << ", " << y4<< std::endl
					//		<< "\t" << x1b << ", " << y1b << "\t" << x2b << ", " << y2b << "\t" << x3b << ", " << y3b << "\t" << x4b << ", " << y4b << std::endl;
					//std::cout << "maximalPoints" << std::endl;
					//for (uInt i = 0; i< maximalPoints.size(); i++)
					//	std::cout << "\t**" << maximalPoints.at(i).x() << ", " << maximalPoints.at(i).y() << std::endl;
					//std::cout << "sourcePoints2" << std::endl;
					//std::cout << sourcePoints2 << std::endl;
					//std::cout << "&&& " << maximalPoints.at(0).x() << ", " << sourcePoints2[0][0] << std::endl;

					/*
					double		x5 = 0, y5 = 0, x6 = 0, y6 = 0, x7 = 0, y7 = 0,
							xc = ((double)cols)/2, yc = ((double)rows)/2;

					linesIntersectionPoint(x1, y1, x2, y2, x3, y3, x4, y4, x5, y5);
					linesIntersectionPoint(x3, y3, x2, y2, x1, y1, x4, y4, x6, y6);
					//getHorizont(x1, y1, x2, y2, x3, y3, x4, y4, x5, y5, x6, y6);
						
					closestPointAtLine(xc, yc, x5, y5,x6, y6,x7, y7);
	*/
					/*myDrawLine(destino, x5, y5, x6, y6);
					QList<QPoint> horizontCenter, imageCenter;
					horizontCenter.append(QPoint(x7, y7));
					imageCenter.append(QPoint(xc, yc));
					drawPoints(horizontCenter, destino);
					drawPoints(imageCenter, destino);*/

					/*const double m1 = (y5-y6)/(x5-x6), m2 = (y7-yc)/(x7-xc);
					std::cout << "m1 " << m1 << std::endl;
					std::cout << "m2 " << m2 << std::endl;
					std::cout << "-m1*m2 " << -((y5-y6)*(y7-yc))/((x5-x6)*(x7-xc)) << std::endl;
					Q_ASSERT(-((y5-y6)*(y7-yc))/((x5-x6)*(x7-xc))==1);
					
					const double	h = NORM(xc-x7, yc-y7),
							a = NORM(x5-x7, y5-y7),
							b = NORM(x6-x7, y6-y7);

					std::cout << "h = " << h << ", a = " << a << ", b = " << b << std::endl;
					if (h*h > a*b)
						std::cout << "FOV = " << sqrt(h*h - a*b) << std::endl;*/

					//////////////////////////////////////////////////////////////////////
					const double	pA[3] = { cols/2, rows, 1},
							pB[3] = { cols/2, rows/distance, 1};
					Vector<>	vA(3), vB(3), vC(3), vD(3), vE(3), temp(3);

					vA = H*Vector<3>(pA);
					vB = H*Vector<3>(pB);

					normalizeHomogeneousCoordinates(vA);
					normalizeHomogeneousCoordinates(vB);

					temp = vB-vA;

					double l = NORM(temp[0], temp[1]);
					temp *= (50/l);

					vC = vB + sqrt(2)*temp;
					vD = vB + RotationMatrix2d(PI/4)*temp;
					vE = vB + RotationMatrix2d(-PI/4)*temp;

					QVImage<uChar,3> wrappedC3 = wrapped;

					//std::cout << "Point " << "\n\t" << vA  << "\n\t" << vB  << "\n\t" << vC  << "\n\t" << vD  << "\n\t" << vE << std::endl;
					QList<QPoint> plist;

					plist.append(QPoint(vA[0] + cols/2, vA[1] + rows/2));
					plist.append(QPoint(vB[0] + cols/2, vB[1] + rows/2));
					plist.append(QPoint(vC[0] + cols/2, vC[1] + rows/2));
					plist.append(QPoint(vD[0] + cols/2, vD[1] + rows/2));
					plist.append(QPoint(vE[0] + cols/2, vE[1] + rows/2));

					drawPoints(plist, wrappedC3);

					/////////////////////////////////////////////

					Matrix <> Hinv(3,3);
					Vector<> vAp(3), vBp(3), vCp(3), vDp(3), vEp(3);
					Hinv = SVD<>(H).get_pinv();

					vAp = Hinv*vA;
					vBp = Hinv*vB;
					vCp = Hinv*vC;
					vDp = Hinv*vD;
					vEp = Hinv*vE;

					normalizeHomogeneousCoordinates(vAp);
					normalizeHomogeneousCoordinates(vBp);
					normalizeHomogeneousCoordinates(vCp);
					normalizeHomogeneousCoordinates(vDp);
					normalizeHomogeneousCoordinates(vEp);

					/////////////////////////////////////////////

					double		x5 = 0, y5 = 0, x6 = 0, y6 = 0, x7 = 0, y7 = 0,
							xc = ((double)cols)/2, yc = ((double)rows)/2;

					linesIntersectionPoint(	vBp[0], vBp[1], vEp[0], vEp[1],		vCp[0], vCp[1], vDp[0], vDp[1],		x5, y5);
					linesIntersectionPoint(	vCp[0], vCp[1], vEp[0], vEp[1],		vBp[0], vBp[1], vDp[0], vDp[1],		x6, y6);
					closestPointAtLine(xc, yc, x5, y5, x6, y6, x7, y7);

					myDrawLine(destino, x5, y5, x6, y6);
					QList<QPoint> horizontCenter, imageCenter;
					horizontCenter.append(QPoint(x7, y7));
					imageCenter.append(QPoint(xc, yc));
					drawPoints(horizontCenter, destino);
					drawPoints(imageCenter, destino);

					///////////////////////////////////////////////7

					QList<QPoint> plist2;
					plist2.append(QPoint(vAp[0], vAp[1]));
					plist2.append(QPoint(vBp[0], vBp[1]));
					plist2.append(QPoint(vCp[0], vCp[1]));
					plist2.append(QPoint(vDp[0], vDp[1]));
					plist2.append(QPoint(vEp[0], vEp[1]));

					//std::cout << "V " << "\n\t" << vP[1] << "\n\t" << Hinv*H*Vector<3>(pA) << std::endl;

					drawPoints(plist2, destino);

					setPropertyValue< QVImage<uChar,3> >("Wrapped", wrappedC3);
					setPropertyValue< QVImage<uChar,1> >("Wrapped2", wrapped2);

					}
				}

			timeFlag("Calibrate");

			// Fin rectificación ////////////////////////
			//drawPoints(maximalPoints, destino);
			setPropertyValue< QVImage<uChar,3> >("Corners", destino);
			timeFlag("Draw corners");
		
			///////////////////////////////////////
			/*QVImage<uChar> buffer;
			QList<QPoint> pointList;
			FindPeaks3x3GetBufferSize(cornerResponseImage, buffer);		
			FindPeaks3x3(cornerResponseImage, buffer, pointList, 0.0, ippiNormInf, 1, hotPoints.size());
			timeFlag("Find Peaks");*/
			}
	};

int main(int argc, char *argv[])
	{
	QVApplication app(argc, argv,
		
		"Example program for QVision library. Applies corner detection over an input video."

		);

	QVMPlayerCamera camera("Video");
	MyWorker worker("Corners Worker");
	camera.link(&worker,"Input image");

	QVGUI interface;

	QVImageCanvas imageCanvas("Corners");
	imageCanvas.linkProperty(worker, "Corners");

	QVImageCanvas imageCanvas2("Wrapped");
	imageCanvas2.linkProperty(worker, "Wrapped");

	QVImageCanvas imageCanvas4("Wrapped 2");
	imageCanvas4.linkProperty(worker, "Wrapped2");

	return app.exec();
	}

/// \endcond


// Solucion en el Mathematica:
//	f = a0 + a1 x + a2 y + a3 x y + a4 x^2 + a5 y^2;
//	Solve[{D[f, x] == 0, D[f, y] == 0} // Simplify, {x, y}] // Simplify
/*bool minimumAtCuadraticInterpolation(const Matrix <> &X, const Vector<> &y, Vector<> &xDest)
	{
	Matrix<> A(X.num_rows(),6);

	for (uInt n = 0; n < X.num_rows(); n++)
		{
		const double	x0 = X[n][0], x1 = X[n][1];
		double		row[6] = {	1,	x0,	x1,	x0*x1,	x0*x0, x1*x1	};
		A[n] = Vector<6>(row);
		}

	Vector<> a = SVD<>(A).get_pinv()*y;

	const double denom = a[3]*a[3] - 4*a[4]*a[5];
	if (denom == 0)
		return false;

	xDest[0] = (-a[2]*a[3]+2*a[1]*a[5]) / denom;
	xDest[1] = (-a[1]*a[3]+2*a[2]*a[4]) / denom;

	return true;
	}

// Solucion en el Mathematica:
//	f = a0 + a1 x + a2 y + a3 x^2 + a4 y^2;
//	Solve[{D[f, x] == 0, D[f, y] == 0} // Simplify, {x, y}] // Simplify
bool minimumAtCuadraticInterpolation2(const Matrix <> &X, const Vector<> &y, Vector<> &xDest)
	{
	Matrix<> A(X.num_rows(),5);

	for (uInt n = 0; n < X.num_rows(); n++)
		{
		const double	x0 = X[n][0], x1 = X[n][1];
		double		row[5] = {	1,	x0,	x1,	x0*x0, x1*x1	};
		A[n] = Vector<5>(row);
		}

	Vector<> a = SVD<>(A).get_pinv()*y;

	xDest[0] = -a[1]/(2*a[3]);
	xDest[1] = -a[2]/(2*a[4]);

	return true;
	}*/

// Given a pixel in an image, this function gets the coordinates of the points and the gray-level value
// in the image of the points in a square neighbourhood of the pixel, of size 2*margin+1
/*bool getMatrices(const QVImage<sFloat> &image, const QPoint p, Matrix<> &X, Vector<> &y, uInt margin)
	{
	const uInt	cols = image.getCols(), rows = image.getRows(),
			col = p.x(), row = p.y();

	if (col < margin || col > cols -margin || row < margin || row > rows -margin)
		return false;

	for (uInt i = col - 2, index = 0; i <= col +2; i++)
		for (uInt j = row - 2; j <= row +2; j++, index++)
			{
			double row[2] = {i, j};
			X[index] = Vector<2>(row);
			y[index] = image(i, j);
			}

	return true;
	}

void getSubpixelPrecissionPoints2(const QVImage<sFloat> &weightImage, QList<QPoint> &pointList, Matrix<> &pointsMatrix)
	{
	for (uInt n = 0; n < maximalPoints.size(); n++)
		{
		Matrix<> X(25,2);
		Vector <> y(25), x0(2);
		getMatrices(cornerResponseImage, maximalPoints.at(n), X, y,0);

		if (minimumAtCuadraticInterpolation2(X, y, x0))
			sourcePoints[n] = x0;
		else	{
			sourcePoints[n][0] = -100;
			sourcePoints[n][1] = -100;
			}
		//std::cout << "Matrix X" << X << std::endl;
		//std::cout << "Matrix y" << y << std::endl;
		
		std::cout << "vector " << maximalPoints.at(n).x() << ", " << maximalPoints.at(n).y() << "\t" << sourcePoints[n] << std::endl;
		}
*/

/*// This function retunrs an integer value that indicates how well the homography maps the source points
// in a list, to the points in the destination list.
uInt	testErrorHomography(const QList<QPoint> &sourcePoints, const QList<QPoint> &destinationPoints, const Matrix<> homography)
	{
	uInt accum = 0;
	for (uInt i = 0; i < sourcePoints.size(); i++)
		accum += (applyHomography(sourcePoints.at(i), homography) -destinationPoints.at(i)).manhattanLength();

	return accum;
	}*/
