/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvgui.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVGUI_H
#define QVGUI_H

#include <QWidget>
#include <QMap>
#include <QCloseEvent>

#include <QToolBox>
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QGroupBox>

#include <qvcore/qvapplication.h>
#include <qvcore/qvimage.h>
#include <qvgui/qvimagecanvas.h>
#include <qvgui/qvparamsinspectorwidget.h>
#include <qvgui/qvcamerawidgetsmall.h>
#include <qvgui/qvworkerinterfacesmall.h>

/*!
@class QVGUI qvgui.h qvgui/qvgui.h
@brief Main window for QVision applications.

This is a class that should be used to create an unique GUI object for any QVision application. Creating this object in a QVision application is optional. If done so, the application will provide a window to the user with different sliders, checkboxes and buttons to control worker input properties (see @ref DynamicProperties ) of integer, double and boolean type, and video input flow for the program:

\image html qvgui_rotoscoper.png

This object should be created after the QVApplication object of the program, and before the call to <i>exec()</i>, generally in the <i>main()</i> function. For a general usage of this class in a real QVision application, see @ref ProgrammingModel section.

\todo Can be created before workers and cameras?

\section QVGUIWidgets QVGUI widgets.
The buttons and sliders are grouped in different widgets in the QVision interface:

\subsection CameraWidget Camera widget.
For every camera object registered in the system there is a widget in the interface, known as the camera widget, that contains buttons to control its input flow. An example of it is depicted below:

\image html qvgui_rotoscoper_camera.png

Follows a description of the buttons and their functionallity:

<table>
<tr><td><b>Pause button</b> <img src="qvgui_pause.png" /></td><td>This button freezes the image that the camera offers to the workers registered in it. If the camera was open in real time mode, the video will continue in the background, but the workers will not read fresh frames. Otherwise the camera will continue exactly at the frame following the one where it was paused, when resuming.
</td></tr>

<tr><td><b>Resume button</b> <img src="qvgui_play.png" /></td><td>This button resumes actualizations of the images that the registered workers read from the cameras.</td></tr>

<tr><td><b>Step button</b> <img src="qvgui_step.png" /></td><td>When the camera is paused, this button makes it read the next frame in the video input, but keeps the camera paused.</td></tr>

<tr><td><b>Stop button</b> <img src="qvgui_stop.png" /></td><td>This button stops the camera from reading frames from the input source.</td></tr>
<tr><td><b>Camera extended controls button</b> <img src="qvgui_cameramaximize.png" /></td><td>This button opens a new window, with extended control functionallity for the camera object, like reproduction speed, and information about the video source, like fps, size of the images, and so:<p />
	<img src="qvgui_cameraextended.png" /></td></tr>
</table>

\subsection WorkerWidget Worker widget.
As for the cameras, every worker object registered in the system has its own widget in the interface. It contains several buttons that control its execution. An example of it is depicted below:

\image html qvgui_rotoscoper_cannyworker.png

You can see a preview plot of cpu usage, and some buttons. Execution for the workers can be paused, resumed, stepped, and stopped with them, as if workers were cameras.

\note CPU stat plot depends on execution time between time flags, so when two or more workers compete for one CPU, times can differ than the used when each worker is run in its own CPU.

Follows a list with the detailed description of the buttons and their functionallity:

<table>
<tr><td><b>Pause button</b> <img src="qvgui_pause.png" /></td><td>When a worker is paused, calls to @ref QVWorker::iterate function are suspended, until the user pushes <i>resume</i> button. Any worker with an input property connected to an output property of a paused worker is also paused, until the latter resumes the execution.</td></tr>

<tr><td><b>Resume button</b> <img src="qvgui_play.png" /></td><td>If the worker is paused, its execution can be resumed by the user pushing this button.</td></tr>

<tr><td><b>Step button</b> <img src="qvgui_step.png" /></td><td>If this button is pushed when a worker is paused, function <i>iterate()</i> will be called one single time, and then worker status will return to <i>paused</i>.</td></tr>

<tr><td><b>Stop button</b> <img src="qvgui_stop.png" /></td><td>This button stops the worker. No more calls to <i>iterate()</i> function will be done, the properties will be unlinked from any other property holder, and freezed, allowing any other worker connected synchrounously to this to continue with those values.</td></tr>

<tr><td><b>Cpu statistics button</b> <img src="qvgui_cpustats.png" /></td><td>This button opens a detailed cpu usage plot. It shows a window with the cpu time stadistical plot of the timeflags defined in the <i>iterate()</i> function of the worker. You can see an example of this plot in the following figure:<p />
	<img src="qvgui_cpustat_canny.png" /></td></tr>
</table>

\subsection CPUFlags CPU usage stadistics.
\todo Here will be explaineda time flags and console and gui stats
\todo explicar funcionamiento de los timeflags y referenciar desde aqui.

\subsection ParamInspectorWidget Param inspector widget.
\todo explicar aqui la inspeccion de parametros en tiempo real.

@ingroup qvgui
*/

class QVGUI: public QWidget
{
	Q_OBJECT
public:	
	//Ui::QVisionForm form;

	/// @brief Constructs a QVGUI object, given the QVision object which work (properties, result images, 
	/// etc...) will be inspected.
	/// @param name identifier for the object to be created.
	QVGUI(QWidget *parent = 0);
	~QVGUI();

	void closeEvent(QCloseEvent *event);

	/// \cond
	// Adds controls for a QVCamera object to the interface, allowing the user to control and inspect the state of
	// the camera.
	// param camera QVCamera object which will be controlled
	void add(QVCamera *camera);

	// Adds a new QVPropertyContainer, which double, integer and boolean properties will be added to the  interface to be
	// controlled in real time by the final user
	// the camera.
	// param name Identification name of the QVPropertyContainer in the interface
	// param object QVPropertyContainer object which property controls will be shown
	void add(QVWorker * worker);
	//bool setVisible(QVWorker *worker,QString prop_orig,float low=0.0,float high=255.0);
	/// \endcond

signals:
	/// @brief Emited when the image canvas window closes.
	///
	void closed();
		
private:
	QHBoxLayout *hboxlayout;
	QWidget *leftwidget;
	QToolBox *toolbox;
	QGroupBox *qgroupbox;
	QVBoxLayout *leftvboxlayout;
	QVBoxLayout *groupboxlayout;
	QList<QVImageCanvas*> qvimagecanvaslist;
};

#endif
